#include "io.h"
#include "io_internal.h"
#include "array.h"
#include "iob.h"
#include <errno.h>
#include "pthread.h"
#include "fmt.h"
#include "buffer.h"
//#include "alloca.h"
#include "stralloc.h"
//#include "dns.h"
#include "ip6.h"
#include "socket.h"
#include "taia.h"
#include "scan.h"
#include "openreadclose.h"
#include "str.h"
#include <semaphore.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/resource.h>
#include <pwd.h>
#include <grp.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <sys/ioctl.h>

int debuglog=0;

typedef unsigned char	cc_t;
typedef unsigned int	speed_t;
typedef unsigned int	tcflag_t;

#define __USE_MISC
#define NCCS 32
struct termios
  {
    tcflag_t c_iflag;		/* input mode flags */
    tcflag_t c_oflag;		/* output mode flags */
    tcflag_t c_cflag;		/* control mode flags */
    tcflag_t c_lflag;		/* local mode flags */
    cc_t c_line;			/* line discipline */
    cc_t c_cc[NCCS];		/* control characters */
    speed_t c_ispeed;		/* input speed */
    speed_t c_ospeed;		/* output speed */
#define _HAVE_STRUCT_TERMIOS_C_ISPEED 1
#define _HAVE_STRUCT_TERMIOS_C_OSPEED 1
  };

/* c_cc characters */
#define VINTR 0
#define VQUIT 1
#define VERASE 2
#define VKILL 3
#define VEOF 4
#define VTIME 5
#define VMIN 6
#define VSWTC 7
#define VSTART 8
#define VSTOP 9
#define VSUSP 10
#define VEOL 11
#define VREPRINT 12
#define VDISCARD 13
#define VWERASE 14
#define VLNEXT 15
#define VEOL2 16

/* c_iflag bits */
#define IGNBRK	0000001
#define BRKINT	0000002
#define IGNPAR	0000004
#define PARMRK	0000010
#define INPCK	0000020
#define ISTRIP	0000040
#define INLCR	0000100
#define IGNCR	0000200
#define ICRNL	0000400
#define IUCLC	0001000
#define IXON	0002000
#define IXANY	0004000
#define IXOFF	0010000
#define IMAXBEL	0020000
#define IUTF8	0040000

/* c_oflag bits */
#define OPOST	0000001
#define OLCUC	0000002
#define ONLCR	0000004
#define OCRNL	0000010
#define ONOCR	0000020
#define ONLRET	0000040
#define OFILL	0000100
#define OFDEL	0000200
#if defined __USE_MISC || defined __USE_XOPEN
# define NLDLY	0000400
# define   NL0	0000000
# define   NL1	0000400
# define CRDLY	0003000
# define   CR0	0000000
# define   CR1	0001000
# define   CR2	0002000
# define   CR3	0003000
# define TABDLY	0014000
# define   TAB0	0000000
# define   TAB1	0004000
# define   TAB2	0010000
# define   TAB3	0014000
# define BSDLY	0020000
# define   BS0	0000000
# define   BS1	0020000
# define FFDLY	0100000
# define   FF0	0000000
# define   FF1	0100000
#endif

#define VTDLY	0040000
#define   VT0	0000000
#define   VT1	0040000

#ifdef __USE_MISC
# define XTABS	0014000
#endif

/* c_cflag bit meaning */
#ifdef __USE_MISC
# define CBAUD	0010017
#endif
#define  B0	0000000		/* hang up */
#define  B50	0000001
#define  B75	0000002
#define  B110	0000003
#define  B134	0000004
#define  B150	0000005
#define  B200	0000006
#define  B300	0000007
#define  B600	0000010
#define  B1200	0000011
#define  B1800	0000012
#define  B2400	0000013
#define  B4800	0000014
#define  B9600	0000015
#define  B19200	0000016
#define  B38400	0000017
#ifdef __USE_MISC
# define EXTA B19200
# define EXTB B38400
#endif
#define CSIZE	0000060
#define   CS5	0000000
#define   CS6	0000020
#define   CS7	0000040
#define   CS8	0000060
#define CSTOPB	0000100
#define CREAD	0000200
#define PARENB	0000400
#define PARODD	0001000
#define HUPCL	0002000
#define CLOCAL	0004000
#ifdef __USE_MISC
# define CBAUDEX 0010000
#endif
#define  B57600   0010001
#define  B115200  0010002
#define  B230400  0010003
#define  B460800  0010004
#define  B500000  0010005
#define  B576000  0010006
#define  B921600  0010007
#define  B1000000 0010010
#define  B1152000 0010011
#define  B1500000 0010012
#define  B2000000 0010013
#define  B2500000 0010014
#define  B3000000 0010015
#define  B3500000 0010016
#define  B4000000 0010017
#define __MAX_BAUD B4000000
#ifdef __USE_MISC
# define CIBAUD	  002003600000		/* input baud rate (not used) */
# define CRTSCTS  020000000000		/* flow control */
#endif

/* c_lflag bits */
#define ISIG	0000001
#define ICANON	0000002
#if defined __USE_MISC || defined __USE_XOPEN
# define XCASE	0000004
#endif
#define ECHO	0000010
#define ECHOE	0000020
#define ECHOK	0000040
#define ECHONL	0000100
#define NOFLSH	0000200
#define TOSTOP	0000400
#ifdef __USE_MISC
# define ECHOCTL 0001000
# define ECHOPRT 0002000
# define ECHOKE	 0004000
# define FLUSHO	 0010000
# define PENDIN	 0040000
#endif
#define IEXTEN	0100000

/* tcflow() and TCXONC use these */
#define	TCOOFF		0
#define	TCOON		1
#define	TCIOFF		2
#define	TCION		3

/* tcflush() and TCFLSH use these */
#define	TCIFLUSH	0
#define	TCOFLUSH	1
#define	TCIOFLUSH	2

/* tcsetattr uses these */
#define	TCSANOW		0
#define	TCSADRAIN	1
#define	TCSAFLUSH	2


#define _IOT_termios /* Hurd ioctl type field.  */ \
  _IOT (_IOTS (cflag_t), 4, _IOTS (cc_t), NCCS, _IOTS (speed_t), 2)

/* ========================================================================= */
/* crc32.h -- tables for rapid CRC calculation
 * Generated automatically by crc32.c
 */

static const uint32 crc_table[256] =
{
    0x00000000UL, 0x77073096UL, 0xee0e612cUL, 0x990951baUL, 0x076dc419UL,
    0x706af48fUL, 0xe963a535UL, 0x9e6495a3UL, 0x0edb8832UL, 0x79dcb8a4UL,
    0xe0d5e91eUL, 0x97d2d988UL, 0x09b64c2bUL, 0x7eb17cbdUL, 0xe7b82d07UL,
    0x90bf1d91UL, 0x1db71064UL, 0x6ab020f2UL, 0xf3b97148UL, 0x84be41deUL,
    0x1adad47dUL, 0x6ddde4ebUL, 0xf4d4b551UL, 0x83d385c7UL, 0x136c9856UL,
    0x646ba8c0UL, 0xfd62f97aUL, 0x8a65c9ecUL, 0x14015c4fUL, 0x63066cd9UL,
    0xfa0f3d63UL, 0x8d080df5UL, 0x3b6e20c8UL, 0x4c69105eUL, 0xd56041e4UL,
    0xa2677172UL, 0x3c03e4d1UL, 0x4b04d447UL, 0xd20d85fdUL, 0xa50ab56bUL,
    0x35b5a8faUL, 0x42b2986cUL, 0xdbbbc9d6UL, 0xacbcf940UL, 0x32d86ce3UL,
    0x45df5c75UL, 0xdcd60dcfUL, 0xabd13d59UL, 0x26d930acUL, 0x51de003aUL,
    0xc8d75180UL, 0xbfd06116UL, 0x21b4f4b5UL, 0x56b3c423UL, 0xcfba9599UL,
    0xb8bda50fUL, 0x2802b89eUL, 0x5f058808UL, 0xc60cd9b2UL, 0xb10be924UL,
    0x2f6f7c87UL, 0x58684c11UL, 0xc1611dabUL, 0xb6662d3dUL, 0x76dc4190UL,
    0x01db7106UL, 0x98d220bcUL, 0xefd5102aUL, 0x71b18589UL, 0x06b6b51fUL,
    0x9fbfe4a5UL, 0xe8b8d433UL, 0x7807c9a2UL, 0x0f00f934UL, 0x9609a88eUL,
    0xe10e9818UL, 0x7f6a0dbbUL, 0x086d3d2dUL, 0x91646c97UL, 0xe6635c01UL,
    0x6b6b51f4UL, 0x1c6c6162UL, 0x856530d8UL, 0xf262004eUL, 0x6c0695edUL,
    0x1b01a57bUL, 0x8208f4c1UL, 0xf50fc457UL, 0x65b0d9c6UL, 0x12b7e950UL,
    0x8bbeb8eaUL, 0xfcb9887cUL, 0x62dd1ddfUL, 0x15da2d49UL, 0x8cd37cf3UL,
    0xfbd44c65UL, 0x4db26158UL, 0x3ab551ceUL, 0xa3bc0074UL, 0xd4bb30e2UL,
    0x4adfa541UL, 0x3dd895d7UL, 0xa4d1c46dUL, 0xd3d6f4fbUL, 0x4369e96aUL,
    0x346ed9fcUL, 0xad678846UL, 0xda60b8d0UL, 0x44042d73UL, 0x33031de5UL,
    0xaa0a4c5fUL, 0xdd0d7cc9UL, 0x5005713cUL, 0x270241aaUL, 0xbe0b1010UL,
    0xc90c2086UL, 0x5768b525UL, 0x206f85b3UL, 0xb966d409UL, 0xce61e49fUL,
    0x5edef90eUL, 0x29d9c998UL, 0xb0d09822UL, 0xc7d7a8b4UL, 0x59b33d17UL,
    0x2eb40d81UL, 0xb7bd5c3bUL, 0xc0ba6cadUL, 0xedb88320UL, 0x9abfb3b6UL,
    0x03b6e20cUL, 0x74b1d29aUL, 0xead54739UL, 0x9dd277afUL, 0x04db2615UL,
    0x73dc1683UL, 0xe3630b12UL, 0x94643b84UL, 0x0d6d6a3eUL, 0x7a6a5aa8UL,
    0xe40ecf0bUL, 0x9309ff9dUL, 0x0a00ae27UL, 0x7d079eb1UL, 0xf00f9344UL,
    0x8708a3d2UL, 0x1e01f268UL, 0x6906c2feUL, 0xf762575dUL, 0x806567cbUL,
    0x196c3671UL, 0x6e6b06e7UL, 0xfed41b76UL, 0x89d32be0UL, 0x10da7a5aUL,
    0x67dd4accUL, 0xf9b9df6fUL, 0x8ebeeff9UL, 0x17b7be43UL, 0x60b08ed5UL,
    0xd6d6a3e8UL, 0xa1d1937eUL, 0x38d8c2c4UL, 0x4fdff252UL, 0xd1bb67f1UL,
    0xa6bc5767UL, 0x3fb506ddUL, 0x48b2364bUL, 0xd80d2bdaUL, 0xaf0a1b4cUL,
    0x36034af6UL, 0x41047a60UL, 0xdf60efc3UL, 0xa867df55UL, 0x316e8eefUL,
    0x4669be79UL, 0xcb61b38cUL, 0xbc66831aUL, 0x256fd2a0UL, 0x5268e236UL,
    0xcc0c7795UL, 0xbb0b4703UL, 0x220216b9UL, 0x5505262fUL, 0xc5ba3bbeUL,
    0xb2bd0b28UL, 0x2bb45a92UL, 0x5cb36a04UL, 0xc2d7ffa7UL, 0xb5d0cf31UL,
    0x2cd99e8bUL, 0x5bdeae1dUL, 0x9b64c2b0UL, 0xec63f226UL, 0x756aa39cUL,
    0x026d930aUL, 0x9c0906a9UL, 0xeb0e363fUL, 0x72076785UL, 0x05005713UL,
    0x95bf4a82UL, 0xe2b87a14UL, 0x7bb12baeUL, 0x0cb61b38UL, 0x92d28e9bUL,
    0xe5d5be0dUL, 0x7cdcefb7UL, 0x0bdbdf21UL, 0x86d3d2d4UL, 0xf1d4e242UL,
    0x68ddb3f8UL, 0x1fda836eUL, 0x81be16cdUL, 0xf6b9265bUL, 0x6fb077e1UL,
    0x18b74777UL, 0x88085ae6UL, 0xff0f6a70UL, 0x66063bcaUL, 0x11010b5cUL,
    0x8f659effUL, 0xf862ae69UL, 0x616bffd3UL, 0x166ccf45UL, 0xa00ae278UL,
    0xd70dd2eeUL, 0x4e048354UL, 0x3903b3c2UL, 0xa7672661UL, 0xd06016f7UL,
    0x4969474dUL, 0x3e6e77dbUL, 0xaed16a4aUL, 0xd9d65adcUL, 0x40df0b66UL,
    0x37d83bf0UL, 0xa9bcae53UL, 0xdebb9ec5UL, 0x47b2cf7fUL, 0x30b5ffe9UL,
    0xbdbdf21cUL, 0xcabac28aUL, 0x53b39330UL, 0x24b4a3a6UL, 0xbad03605UL,
    0xcdd70693UL, 0x54de5729UL, 0x23d967bfUL, 0xb3667a2eUL, 0xc4614ab8UL,
    0x5d681b02UL, 0x2a6f2b94UL, 0xb40bbe37UL, 0xc30c8ea1UL, 0x5a05df1bUL,
    0x2d02ef8dUL
};
#define DO1 crc = crc_table[((int)crc ^ (*buf++)) & 0xff] ^ (crc >> 8)
#define DO8 DO1; DO1; DO1; DO1; DO1; DO1; DO1; DO1

/* ========================================================================= */
static uint32 crc32(uint32 crc, const unsigned char *buf, uint64 len) {
    if (!buf) return 0;

    crc = crc ^ 0xffffffffUL;
    for (;len >= 8;) {
        DO8;
        len -= 8;
    }
    for (;len;--len) {
        DO1;
    }
    return crc ^ 0xffffffffUL;
}

int randfd=-1;

enum conntype {
  KEYBOARD,
  COMPORT
};

typedef struct _serialdevice_ serialdevice;

struct _serialdevice_ {
  enum conntype t;
  int64 fd;
  array r;
  io_batch iob;
  stralloc command;
  stralloc send_packet;
  char* buf;
  int len;
  unsigned long long received,sent;
  stralloc devicename;
  stralloc nickname;
  struct termios ot,nt;
  int status; // 0=reading/waiting 1=writing
};

static void stralloc_reset(stralloc *sa) {
  if (sa->s) free(sa->s);
  sa->s=0;
  sa->len=sa->a=0;
}

static void serial_init(serialdevice *sd) {
  if (!sd) return;
  byte_zero(sd,sizeof(serialdevice));
  sd->fd=-1;
}

static int serial_close(serialdevice *sd) {
  if (!sd || (sd->fd==-1)) return 0;
  io_close(sd->fd);
  stralloc_reset(&sd->nickname);
  stralloc_reset(&sd->devicename);
  return 1;
}

//serial_init should be called first
static int serial_open(serialdevice *sd, char *name, char *nick) {
  if (!name || !sd || (sd->fd!=-1)) return 0;
  stralloc_init(&sd->devicename);
  stralloc_init(&sd->nickname);
  if (!stralloc_copys(&sd->devicename,name)) return 0;
  if (!stralloc_copys(&sd->nickname,nick)) goto nick_failed;
  sd->fd=open(name,O_RDWR|O_NOCTTY|O_NONBLOCK);
  if (sd->fd==-1) {
    stralloc_reset(&sd->nickname);
nick_failed:
    stralloc_reset(&sd->devicename);
    return 0;
  }
  if (io_fd(sd->fd)) {
    ioctl(sd->fd, TCGETS, &sd->nt);
    io_setcookie(sd->fd,sd);
    io_nonblock(sd->fd);
    io_wantread(sd->fd);
    io_dontwantwrite(sd->fd);
    sd->t=COMPORT;
    return 1;
  }
  serial_close(sd);
  return 0;
}

static int serial_conf(serialdevice *sd) {
  if (!sd || (sd->fd==-1)) return 0;
  byte_copy(&sd->ot,sizeof(sd->ot),&sd->nt);
  sd->ot.c_iflag&=~(IXON|IXOFF|IGNBRK|INPCK|IGNCR|INLCR|ICRNL|IUCLC|IXANY|IMAXBEL);
  sd->ot.c_iflag|=(BRKINT);
  sd->ot.c_oflag&=~(OPOST|OLCUC|ONLCR|OCRNL|ONOCR|ONLRET|OFILL|OFDEL|NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY);
//  sd->ot.c_oflag|=();
  sd->ot.c_cflag&=~(CBAUD|CSIZE|PARENB|CRTSCTS);
  sd->ot.c_cflag|=(B115200|CS8|CSTOPB/*|CREAD*/|CLOCAL);
  sd->ot.c_lflag&=~(ISIG|ICANON|ECHO|ECHOE|ECHOK|ECHONL|ECHOCTL|ECHOPRT|ECHOKE|TOSTOP|IEXTEN);
  sd->ot.c_lflag|=(NOFLSH);
  sd->ot.c_cc[VMIN]=0;
  sd->ot.c_cc[VTIME]=0;
  if (ioctl(sd->fd, TCSETS, &sd->ot)==-1) return 0;
  if (ioctl(sd->fd, TCGETS, &sd->nt)==-1) return 0;
  return 1;
}

static serialdevice mycom;
//static serialdevice com2;


static void buffer_putserialconf(buffer *buf, serialdevice *sd) {
  buffer_putsa(buffer_1, &sd->nickname);
  buffer_puts(buf,"-STATUS:\nc_iflag=0x");
  buffer_putxlong(buf, sd->nt.c_iflag);
  buffer_puts(buf,"\nc_oflag=0x");
  buffer_putxlong(buf, sd->nt.c_oflag);
  buffer_puts(buf,"\nc_cflag=0x");
  buffer_putxlong(buf, sd->nt.c_cflag);
  buffer_puts(buf,"\nc_lflag=0x");
  buffer_putxlong(buf, sd->nt.c_lflag);
  int i;
  unsigned char *ch;
  ch=sd->nt.c_cc;
  for (i=0; i<NCCS; ++i,++ch) {
    int j;
    j=*ch;
    buffer_puts(buf, "\nc_cc[");
    buffer_putlong(buf, i);
    buffer_puts(buf, "]=");
    buffer_putxlong(buf, j);
  }
  buffer_putnlflush(buf);
}

unsigned long timeout_secs=1;
tai6464 now,next,nextw,notimeout;

uint64 packet_ok, packet_timeout, packet_timeout2;

static volatile int fini;

void sighandler(int sig) {
  fini=(sig==SIGINT?1:2);	/* 2 for SIGHUP */
}

unsigned short myaddress=-1;

static int header_complete(serialdevice* r) {
  long i;
  long l=array_bytes(&r->r);
  const char* c=array_start(&r->r);
  if (r->t==KEYBOARD)
  {
    for (i=0; i+1<l; ++i) {
      if (c[i]=='\n' && c[i+1]=='\n')
	return i+2;
      if (i+3<l &&
	  c[i]=='\r' && c[i+1]=='\n' &&
	  c[i+2]=='\r' && c[i+3]=='\n')
	return i+4;
    }
  }
  return 0;
}



tai6464 lastevent;
static void handle_read_misc(int64 i,serialdevice* c) {
  char buf[8192];
  int l;
  tai6464 start,end,diffe,diffa;
  taia_now(&start);
  if (debuglog) buffer_putsflush(buffer_1,".");
  taia_sub(&diffe,&start,&lastevent);
  l=io_tryread(i,buf,sizeof buf);
  if (l<=0 && errno==EAGAIN) {
    io_eagain(i);
    return;
  }
  if (l==-3) {
    char a[FMT_ULONG];
    a[fmt_ulong(a,i)]=0;
    buffer_putmflush(buffer_1,"io_error ",a," ",strerror(errno),"\nclose/readerr ",a,"\n");
    io_close(i);
  } else if (l==0) {
    char a[FMT_ULONG];
    char e[FMT_ULONG];
    a[fmt_ulong(a,i)]=0;
    e[fmt_ulong(e,errno)]=0;
    buffer_putmflush(buffer_1, "close/read0 fd=", a, "errno=", e, " error=", strerror(errno), "\n");
    io_close(i);
  } else if (l>0) {
  taia_now(&end);
  taia_sub(&diffa,&end,&start);
  byte_copy(&lastevent,sizeof(start),&start);
  if (debuglog) {
  buffer_puts(buffer_1, "ediff=");
  buffer_putulong(buffer_1, diffe.sec.x);
  buffer_puts(buffer_1, ",");
  buffer_putulong(buffer_1, diffe.nano);
  buffer_putnlflush(buffer_1);
  buffer_puts(buffer_1, "rdiff=");
  buffer_putulong(buffer_1, diffa.sec.x);
  buffer_puts(buffer_1, ",");
  buffer_putulong(buffer_1, diffa.nano);
  buffer_putnlflush(buffer_1);
  }
    /* successfully read some data (l bytes) */
    c->received+=l;
    {
      /* received a request */
      array_catb(&c->r,buf,l);
      if (array_failed(&c->r)) {
	buffer_putsflush(buffer_2, "\n\nArray failed!\n\n");
      } else {
	if (!i) {
	  serialdevice *mom;
	  mom=&mycom;

	  l=header_complete(c);
	  if (!l) return;

	  char* m=array_start(&c->r);
	  switch (*m) {
	    case 'r':
	    {
	      int h1;
	      uint32 datalen, crc=0;
	      uint32* crcp;
	      unsigned short toaddress;
	      //Zufallspaket senden
	      ++m;
	      for (;*m==' ';) ++m;
	      h1=scan_ushort(m,&toaddress);
	      if (!h1) goto usage;
	      m+=h1;

	      for (;*m==' ';) ++m;
	      h1=scan_uint(m,&datalen);
	      if (!h1) goto usage;
	      m+=h1;

	      mom->len=datalen+4+2+2+1+1; //Daten+CRC+Absender+Empfnger+Paketlnge+Mllbyte
//Paketaufbau:
//0: DB l=Lnge des folgenden Pakets
//1: DB 'M'
//2: DW Absender
//4: DW Empfnger
//6: l-8 Byte Daten
//l-3: DD CRC32
	      if (mom->buf) {
		free(mom->buf);
		mom->buf=0;
	      }
	      char* p=mom->buf=malloc(mom->len);
	      unsigned short *s;
	      *p=mom->len-1;
	      ++p;
	      *p='M';
	      ++p;
	      s=(void*)p;
	      *s=myaddress;
	      ++s;
	      *s=toaddress;
	      ++s;
	      p=(void*)s;
	      read(randfd,p,datalen);
	      crc=crc32(crc,(unsigned char*)p,datalen);
	      p+=datalen;
	      crcp=(void*)p;
	      *crcp=crc;
	      if (debuglog) { 
	        buffer_puts(buffer_1, "\n  PacketCRC=");
	        buffer_putulong(buffer_1, *crcp);
	        buffer_putnlflush(buffer_1);
		}
	      break;
	    }
	    default:
usage:
	      buffer_putsflush(buffer_2, "Befehl nicht in Ordnung! Puffer komplett gelscht!\n");
	      array_reset(&c->r);
	      return;
	  }
	  m=array_start(&c->r);
	  long alen=array_bytes(&c->r);
	  byte_copy(m,alen-l,m+l);
	  array_truncate(&c->r,1,alen-l);

	  iob_addbuf(&mom->iob,mom->buf,mom->len);
	  io_dontwantwrite(mom->fd);
	  io_dontwantread(mom->fd);
	  io_wantwrite(mom->fd);
	  io_wantread(mom->fd);
	  io_timeout(mom->fd, nextw);
	  mom->status=1;
	  if (debuglog) { buffer_putm(buffer_1, "\nT->");
	  buffer_putsa(buffer_1, &mom->nickname);
	  buffer_puts(buffer_1, ": \"disabled");
//	  buffer_put(buffer_1, mom->buf+1, mom->len-1);
	  buffer_putsflush(buffer_1, "\" L=");
	  buffer_putlong(buffer_1, mom->len);
	  buffer_putsflush(buffer_1, "\n");
}
	} else {
	  serialdevice *mom;
	  int len,blen;
	  unsigned char *buf;
	  mom=&mycom;
	  mom->status=0;
	  io_timeout(mom->fd, next);
	  len=array_bytes(&c->r);
	  buf=array_start(&c->r);
	  if (debuglog) {
	  buffer_putm(buffer_1, "\nFUNK->");
	  buffer_putsa(buffer_1, &c->nickname);
	  buffer_puts(buffer_1, ": \"disabled");
//	  buffer_put(buffer_1, buf+1, len-1);
	  buffer_puts(buffer_1, "\" PL=");
	  buffer_putlong(buffer_1, buf[0]);
	  buffer_puts(buffer_1, " RL=");
	  buffer_putlong(buffer_1, len);
	  buffer_putsflush(buffer_1, "\n");
	}
	  if ((len>=2) && (buf[1]!='M')) {//Paketstart nicht ok
	    if (debuglog) buffer_putsflush(buffer_1,"Kein Paketanfang gefunden! Puffer gelscht.\n");
	    array_reset(&c->r);
	    return;
	  }

	  if (debuglog) buffer_putsflush(buffer_1,"Paketanfang gefunden!\n");
	  blen=1+buf[0];
	  if (blen<=len) { // packet complete
	    uint32 datalen, crc=0, crcok=0;
	    uint32* crcp;
	    unsigned char* p=buf;
	    unsigned short *s, from, to;
	    if (debuglog) buffer_putsflush(buffer_1, "Paket:\n  Lnge=");
	    datalen=*p-(4+2+2+1);
	    if (debuglog) buffer_putulong(buffer_1, datalen);
	    if (debuglog) buffer_putsflush(buffer_1, "\n  Absender=");
	    ++p;++p;
	    s=(void*)p;
	    from=*s;
	    if (debuglog) buffer_putulong(buffer_1, from);
	    if (debuglog) buffer_putsflush(buffer_1, "\n  Empfnger=");
	    ++s;
	    to=*s;
	    if (debuglog) buffer_putulong(buffer_1, to);
	    ++s;
	    p=(void*)s;
	    crc=crc32(crc,(unsigned char*)p,datalen);
	    p+=datalen;
	    crcp=(void*)p;
	    if (*crcp==crc) {
	      if (debuglog) buffer_putsflush(buffer_1, "\n  CRC ok.\n");
	      crcok=1;
	    } else {
	      if (debuglog) buffer_putsflush(buffer_1, "\n  CRC nicht ok.\n");
	    }
	    if (debuglog) {buffer_puts(buffer_1, "\n  PacketCRC=");
	    buffer_putulong(buffer_1, *crcp);
	    buffer_puts(buffer_1, "\n  MyCRC=");
	    buffer_putulong(buffer_1, crc);
	    buffer_putnlflush(buffer_1);
}
/*	    mom->len=buf[0]+1;
	    mom->buf=malloc(mom->len);
	    byte_copy(mom->buf,mom->len,buf);
	    iob_addbuf_free(&mom->iob,mom->buf,mom->len);
	    io_dontwantwrite(mom->fd);
	    io_dontwantread(mom->fd);
	    io_wantwrite(mom->fd);
	    io_wantread(mom->fd);
	    buffer_putm(buffer_1, "\n");
	    buffer_putsa(buffer_1, &c->nickname);
	    buffer_putm(buffer_1, "->");
	    buffer_putsa(buffer_1, &mom->nickname);
	    buffer_puts(buffer_1, ": \"");
	    buffer_put(buffer_1, mom->buf+1, mom->len-1);
	    buffer_putsflush(buffer_1, "\" L=");
	    buffer_putlong(buffer_1, mom->buf[0]);
	    buffer_putsflush(buffer_1, "\n");*/

	    if (to==myaddress && crcok) {
	      if (debuglog) buffer_puts(buffer_1, "Das Paket war fr uns, also antworten wir mal mit einem gleichlangen Paket.\n");
	      mom->len=datalen+4+2+2+1+1; //Daten+CRC+Absender+Empfnger+Paketlnge+Mllbyte
//Paketaufbau:
//0: DB l=Lnge des folgenden Pakets
//1: DB 'M'
//2: DW Absender
//4: DW Empfnger
//6: l-8 Byte Daten
//l-3: DD CRC32
	      if (mom->buf) {
		free(mom->buf);
		mom->buf=0;
	      }
	      char* p=mom->buf=malloc(mom->len);
	      unsigned short *s;
	      *p=mom->len-1;
	      ++p;
	      *p='M';
	      ++p;
	      s=(void*)p;
	      *s=myaddress;
	      ++s;
	      *s=from;
	      ++s;
	      p=(void*)s;
	      read(randfd,p,datalen);
	      crc=0;
	      crc=crc32(crc,(unsigned char*)p,datalen);
	      p+=datalen;
	      crcp=(void*)p;
	      *crcp=crc;
	      if (debuglog) {buffer_puts(buffer_1, "\n  PacketCRC=");
	      buffer_putulong(buffer_1, *crcp);
	      buffer_putnlflush(buffer_1);
}
	      iob_addbuf(&mom->iob,mom->buf,mom->len);
	      io_dontwantwrite(mom->fd);
	      io_dontwantread(mom->fd);
	      io_wantwrite(mom->fd);
	      io_wantread(mom->fd);
	      io_timeout(mom->fd, nextw);
	      ++packet_ok;
	      mom->status=1;
	    }

	    char* m=array_start(&c->r);
	    long alen=array_bytes(&c->r);
	    if (alen>=blen) {
	      byte_copy(m,alen-blen,m+blen);
	      array_truncate(&c->r,1,alen-blen);
	    } else {
	      array_reset(&c->r);
	    }
	  } else return;
	}
//	array_reset(&c->r);
      }
    }
  }
}


static void handle_write_misc(int64 i, serialdevice *c) {
  int64 r;
  tai6464 start,end,diffe,diffa;
  taia_now(&start);
  if (debuglog) buffer_putsflush(buffer_1,",");
  taia_sub(&diffe,&start,&lastevent);
  r=iob_send(i,&c->iob);
  taia_now(&end);
  taia_sub(&diffa,&end,&start);
  if (r==-1)
    io_eagain(i);
  else if (r<=0) {
    if (r==-3) {
      char a[FMT_ULONG];
      a[fmt_ulong(a,i)]=0;
      buffer_putmflush(buffer_1,"socket_error ",a," ",strerror(errno),"\nclose/writefail ",a,"\n");
      io_close(i);
    } else {	/* returned 0, i.e. we wrote it all */
  byte_copy(&lastevent,sizeof(start),&start);
  if (debuglog) {buffer_puts(buffer_1, "ediff=");
  buffer_putulong(buffer_1, diffe.sec.x);
  buffer_puts(buffer_1, ",");
  buffer_putulong(buffer_1, diffe.nano);
  buffer_putnlflush(buffer_1);
  buffer_puts(buffer_1, "wdiff=");
  buffer_putulong(buffer_1, diffa.sec.x);
  buffer_puts(buffer_1, ",");
  buffer_putulong(buffer_1, diffa.nano);
  buffer_putnlflush(buffer_1);
}
      if (c->t == COMPORT) {
	if (debuglog) {buffer_puts(buffer_1,"packet sent to ");
	buffer_putsa(buffer_1,&c->nickname);
	buffer_puts(buffer_1," fd=");
	buffer_putulong(buffer_1,i);
	buffer_puts(buffer_1," rec=");
	buffer_putulonglong(buffer_1,c->received);
	buffer_puts(buffer_1," sent=");
	buffer_putulonglong(buffer_1,c->sent);
	buffer_putnlflush(buffer_1);
}
//	h->received=h->sent=0;
      }
//      c->buf=0;
      iob_reset(&c->iob);
      io_dontwantwrite(i);
    }
  } else {
    c->sent+=r;
  }
}

//struct termios
//  {
//    tcflag_t c_iflag;		/* input mode flags */
//    tcflag_t c_oflag;		/* output mode flags */
//    tcflag_t c_cflag;		/* control mode flags */
//    tcflag_t c_lflag;		/* local mode flags */
//    cc_t c_line;			/* line discipline */
//    cc_t c_cc[NCCS];		/* control characters */
//    speed_t c_ispeed;		/* input speed */
//    speed_t c_ospeed;		/* output speed */
//#define _HAVE_STRUCT_TERMIOS_C_ISPEED 1
//#define _HAVE_STRUCT_TERMIOS_C_OSPEED 1
//  };
serialdevice tastatur;
int main(int argc,char* argv[],char* envp[]) {
  (void)envp;
  (void)argc;
  (void)argv;
  char* devicename="/dev/ttyS0";
  char* nickname="COM1";
  tai6464 last,tick;
  serial_init(&tastatur);
  tastatur.fd=0;
  tastatur.t=KEYBOARD;
  serial_init(&mycom);

  signal(SIGPIPE,SIG_IGN);

  {
    struct sigaction sa;
    byte_zero(&sa,sizeof(sa));
    sigemptyset(&sa.sa_mask);
    sa.sa_handler=sighandler;
    sigaction(SIGINT,&sa,0);
    sigaction(SIGHUP,&sa,0);
  }

#if !defined(__linux__)
  optind=1;
#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__)
  optreset=1;
#endif
#endif

  for (;;) {
    int i;
    int c=getopt(argc,argv,"d:n:i:h");
    if (c==-1) break;
    switch (c) {
    case 'i':
      i=scan_ushort(optarg,&myaddress);
      if (i==0) {
	buffer_puts(buffer_2,"warning: could not parse address ");
	buffer_puts(buffer_2,optarg+i+1);
	buffer_putsflush(buffer_2,".\n");
	myaddress=-1;
      }
      break;
    case 'd':
      devicename=optarg;
      break;
    case 'n':
      nickname=optarg;
      break;
    default:
    case '?':
    case 'h':
usage:
      buffer_putsflush(buffer_2,
		  "usage: serialtest [-h] -i bind-to-address [-d devicename] [-n nickname]\n"
		  "\n"
		  "\t-h\tprint this help\n"
		  "\t-d\tdevicename (default=/dev/ttyS0)\n"
		  "\t-n\tnickname for device (default=COM1)\n"
		  );
      return 0;
    }
  }

  if (myaddress==(unsigned short)-1) goto usage;

  randfd=open("/dev/urandom",O_RDONLY);

  buffer_putmflush(buffer_1, "Opening ",nickname,"=", devicename, "...");
  if (!serial_open(&mycom,devicename, nickname)) {
    buffer_putsflush(buffer_1, "failed.\n");
    return 1;
  }
  buffer_putsflush(buffer_1, "ok.\n");
  
  buffer_putserialconf(buffer_1, &mycom);
  buffer_putnlflush(buffer_1);

  serial_conf(&mycom);

  if (!byte_diff(&mycom.ot,sizeof(mycom.ot),&mycom.nt)) {
	buffer_putmflush(buffer_1, nickname, " CONF ok.\n");
  } else {
	buffer_putmflush(buffer_1, nickname, " CONF not ok.\n");
  }

  buffer_putserialconf(buffer_1, &mycom);
  buffer_putnlflush(buffer_1);

  io_fd(0);
  io_wantread(0);
  io_setcookie(0,&tastatur);

  buffer_putsflush(buffer_1, "Starting main loop...\n\n");
  for (;;) {
    int64 i;

    if (fini) {
      buffer_putsflush(buffer_1,"closing\n");
      break;
    }

/*    handle_read_misc(tastatur.fd,&tastatur);
    handle_read_misc(com1.fd,&com1);
  io_entry* e=array_get(&io_fds,sizeof(io_entry),com1.fd);
    if (e && e->wantwrite) handle_write_misc(com1.fd,&com1);
*/
    if (timeout_secs)
      io_waituntil(tick);
    else
      io_wait();

    buffer_puts(buffer_1, "\n\nSTATUS: packet_ok=");
    buffer_putulong(buffer_1, packet_ok);
    buffer_puts(buffer_1, "; packet_timeout=");
    buffer_putulong(buffer_1, packet_timeout);
    buffer_puts(buffer_1, "; packet_timeout2=");
    buffer_putulong(buffer_1, packet_timeout2);
    buffer_putsflush(buffer_1, ";\n\n");

    taia_now(&now);
    if (timeout_secs) {
      if (now.sec.x != last.sec.x) {
	byte_copy(&last,sizeof(now),&now);
	byte_copy(&next,sizeof(now),&now);
	next.sec.x += timeout_secs;
	byte_copy(&nextw,sizeof(now),&now);
	nextw.sec.x += timeout_secs*2;
	byte_copy(&tick,sizeof(next),&now);
	++tick.sec.x;

	while ((i=io_timeouted())!=-1) {
	  serialdevice* c=io_getcookie(i);
	  iob_reset(&c->iob);
	  array_reset(&c->r);
	  if (!c->status || !c->buf) {
	    ++packet_timeout2;
	    buffer_putsflush(buffer_1, "TIMEOUT -> no packet present");
	    io_dontwantwrite(c->fd);
	    io_dontwantread(c->fd);
	    io_wantread(c->fd);
	    io_timeout(c->fd, notimeout);
	  } else {
	    ++packet_timeout;
	    buffer_putsflush(buffer_1, "TIMEOUT -> packet resent");
	    iob_addbuf(&c->iob, c->buf, c->len);
	    io_dontwantwrite(c->fd);
	    io_dontwantread(c->fd);
	    io_wantwrite(c->fd);
	    io_wantread(c->fd);
	    io_timeout(c->fd, next);
	  }
	}
      }
    }


   /* HANDLE READ EVENTS */
   while ((i=io_canread())!=-1) {
     serialdevice* c=io_getcookie(i);
     if (!c) {
	char a[FMT_ULONG];
	a[fmt_ulong(a,i)]=0;
	buffer_putmflush(buffer_2,"canthappen ",a,": got read event on socket with no cookie!\n");
	io_dontwantread(i);
	io_close(i);
	continue;
      }

      handle_read_misc(i,c);
    }

    /* HANDLE WRITABLE EVENTS */
    while ((i=io_canwrite())!=-1) {
      serialdevice* c=io_getcookie(i);
      if (!c) {
	char a[FMT_ULONG];
	a[fmt_ulong(a,i)]=0;
	buffer_putmflush(buffer_2,"canthappen ",a,": got write event on socket with no cookie!\n");
	io_dontwantwrite(i);
	io_close(i);
	continue;
      }

      handle_write_misc(i,c);
    }

  }

  serial_close(&mycom);
  close(randfd);
//  serial_close(&com2);
  io_finishandshutdown();
  return 0;
}

/* For the emacs weenies in the crowd.
Local Variables:
   c-basic-offset: 2
   indent-tabs-mode: nil
   c-indent-level: 2
   c-syntactic-indentation: nil
End:
*/
